# cython: cdivision=True
# cython: boundscheck=False
# cython: wraparound=False
# cython: language_level=3
# cython: initializedcheck=False

# Authors: Isak Samsten
# License: BSD 3 clause

import numpy as np

from libc.math cimport M_PI, cos, log, log2, sin, sqrt, INFINITY
from libc.stdlib cimport realloc, malloc, free
from libc.string cimport memset

from ..utils cimport TSArray

cdef void _heap_shift_down(
    HeapElement *heap, Py_ssize_t startpos, Py_ssize_t pos
) noexcept nogil:
    cdef HeapElement newelement = heap[pos]
    cdef Py_ssize_t parentpos
    while pos > startpos:
        parentpos = (pos - 1) >> 1  # // 2
        if newelement.value > heap[parentpos].value:
            heap[pos] = heap[parentpos]
            pos = parentpos
            continue
        break

    heap[pos] = newelement

cdef void _heap_shift_up(
    HeapElement *heap, Py_ssize_t pos, Py_ssize_t endpos
) noexcept nogil:
    cdef Py_ssize_t startpos = pos
    cdef HeapElement newelement = heap[pos]
    cdef Py_ssize_t childpos = 2 * pos + 1
    cdef Py_ssize_t rightpos
    while childpos < endpos:
        rightpos = childpos + 1
        if rightpos < endpos and heap[childpos].value < heap[rightpos].value:
            childpos = rightpos

        heap[pos] = heap[childpos]
        pos = childpos
        childpos = 2 * pos + 1

    heap[pos] = newelement
    _heap_shift_down(heap, startpos, pos)

# A Heap implementation that retains the `k` smallest elements.
# The first element of the heap is the largest of the `k` smallest
# elements and can be retreived in O(1) time. We maintain a separate
# datastructure for the minimum value so that can also be retreived
# in constant time.
#
# `_min` returns the smallest value among k smallest elements.
# `_max` returns the largest value among the top k elements.
#
# The data structure efficently act as a proxy for `argpartition` in numpy.
cdef class Heap:

    def __cinit__(self, Py_ssize_t max_elements):
        self.heap = <HeapElement*> malloc(sizeof(HeapElement) * max_elements)
        self.n_elements = 0
        self.max_elements = max_elements

    def __dealloc__(self):
        if self.heap != NULL:
            free(self.heap)
            self.heap = NULL

    cdef void push(self, Py_ssize_t index, double value) noexcept nogil:
        cdef HeapElement element
        element.index = index
        element.value = value

        if self.n_elements == 0:
            self.heap[0] = element
            self.n_elements += 1
        else:
            if self.n_elements < self.max_elements:
                self.heap[self.n_elements] = element
                self.n_elements += 1
                _heap_shift_down(self.heap, 0, self.n_elements - 1)
            elif self.heap[0].value > element.value:
                self.heap[0] = element
                _heap_shift_up(self.heap, 0, self.max_elements)

    cdef HeapElement maxelement(self) noexcept nogil:
        return self.heap[0]

    cdef double maxvalue(self) noexcept nogil:
        return self.heap[0].value

    cdef HeapElement getelement(self, Py_ssize_t i) noexcept nogil:
        return self.heap[i]

    cdef void reset(self) noexcept nogil:
        self.n_elements = 0

    cdef bint isempty(self) noexcept nogil:
        return self.n_elements == 0

    cdef bint isfull(self) noexcept nogil:
        return self.n_elements == self.max_elements


cdef extern from "Python.h":
    cdef void* PyList_GET_ITEM(list, Py_ssize_t index) nogil

cdef class List:

    def __init__(self, list py_list):
        self.py_list = py_list
        self.size = len(py_list)

    cdef void* get(self, Py_ssize_t i) noexcept nogil:
        return PyList_GET_ITEM(self.py_list, i)


cdef inline void argsort(
    double_or_int *values, Py_ssize_t *samples, Py_ssize_t n
) noexcept nogil:
    if n == 0:
        return
    cdef Py_ssize_t maxd = 2 * <Py_ssize_t> log2(n)
    introsort(values, samples, n, maxd)

cdef inline void swap(
    double_or_int *values,
    Py_ssize_t *samples,
    Py_ssize_t i,
    Py_ssize_t j
) noexcept nogil:
    values[i], values[j] = values[j], values[i]
    samples[i], samples[j] = samples[j], samples[i]

cdef inline double_or_int median3(double_or_int *values, Py_ssize_t n) noexcept nogil:
    cdef double_or_int a = values[0]
    cdef double_or_int b = values[n / 2]
    cdef double_or_int c = values[n - 1]
    if a < b:
        if b < c:
            return b
        elif a < c:
            return c
        else:
            return a
    elif b < c:
        if a < c:
            return a
        else:
            return c
    else:
        return b

cdef void introsort(
    double_or_int *values,
    Py_ssize_t *samples,
    Py_ssize_t n,
    Py_ssize_t maxd
) noexcept nogil:
    cdef double_or_int pivot, value
    cdef Py_ssize_t i, l, r

    while n > 1:
        if maxd <= 0:
            heapsort(values, samples, n)
            return
        maxd -= 1

        pivot = median3(values, n)

        i = l = 0
        r = n
        while i < r:
            value = values[i]
            if value < pivot:
                swap(values, samples, i, l)
                i += 1
                l += 1
            elif value > pivot:
                r -= 1
                swap(values, samples, i, r)
            else:
                i += 1

        introsort(values, samples, l, maxd)
        values += r
        samples += r
        n -= r

cdef inline void sift_down(double_or_int *values, Py_ssize_t *samples,
                           Py_ssize_t start, Py_ssize_t end) noexcept nogil:
    cdef Py_ssize_t child, maxind, root
    root = start
    while True:
        child = root * 2 + 1
        maxind = root
        if child < end and values[maxind] < values[child]:
            maxind = child
        if child + 1 < end and values[maxind] < values[child + 1]:
            maxind = child + 1

        if maxind == root:
            break
        else:
            swap(values, samples, root, maxind)
            root = maxind

cdef void heapsort(
    double_or_int *values, Py_ssize_t *samples, Py_ssize_t n
) noexcept nogil:
    cdef Py_ssize_t start, end

    start = (n - 2) / 2
    end = n
    while True:
        sift_down(values, samples, start, end)
        if start == 0:
            break
        start -= 1

    end = n - 1
    while end > 0:
        swap(values, samples, 0, end)
        sift_down(values, samples, 0, end)
        end = end - 1

cdef int realloc_array(
    void** ptr, Py_ssize_t old_size, Py_ssize_t ptr_size, Py_ssize_t *capacity
)  except -1 nogil:
    cdef void *tmp = ptr[0]
    if old_size >= capacity[0]:
        capacity[0] = old_size * 2
        tmp = realloc(ptr[0], ptr_size * capacity[0])
        if tmp == NULL:
            return -1
    ptr[0] = tmp
    return 0

cdef int safe_realloc(void** ptr, Py_ssize_t new_size) except -1 nogil:
    cdef void *tmp = ptr[0]
    tmp = realloc(ptr[0], new_size)
    if tmp == NULL:
        return -1

    ptr[0] = tmp
    return 0

cdef object to_ndarray_int(Py_ssize_t *arr, Py_ssize_t n):
    cdef Py_ssize_t i
    cdef Py_ssize_t[:] out = np.zeros(n, dtype=np.intp)
    for i in range(n):
        out[i] = arr[i]

    return out.base

cdef object to_ndarray_double(double *arr, Py_ssize_t n):
    cdef Py_ssize_t i
    cdef double[:] out = np.zeros(n, dtype=float)
    for i in range(n):
        out[i] = arr[i]

    return out.base


def _test_ts_array(TSArray arr):
    return arr[0, 0, 0]



# Authors: Isak Samsten
# License: BSD 3 clause

"""
DTW alignment and distance computations.

The :mod:`wildboar.distance.dtw` module implements several functions for
computing DTW alignments and distances.
"""

import math
import numbers
from functools import partial

import numpy as np
from sklearn.utils import check_random_state, check_scalar
from sklearn.utils.validation import _check_sample_weight, _is_arraylike

from ..utils.validation import check_array
from ._distance import pairwise_distance
from ._elastic import _dtw_alignment, _dtw_envelop, _dtw_lb_keogh

__all__ = [
    "dtw_alignment",
    "wdtw_alignment",
    "dtw_distance",
    "wdtw_distance",
    "ddtw_distance",
    "wddtw_distance",
    "dtw_mapping",
    "dtw_envelop",
    "dtw_lb_keogh",
    "jeong_weight",
    "dtw_average",
]


def _compute_warp_size(x_size, r, *, y_size=0):
    check_scalar(r, "r", numbers.Real, min_val=0, max_val=1)
    return max(math.floor(max(x_size, y_size) * r), 1)


def dtw_distance(x, y, *, r=1.0):
    """
    Compute the dynamic time warping distance.

    Parameters
    ----------
    x : array-like of shape (x_timestep, )
        The first time series.
    y : array-like of shape (y_timestep, )
        The second time series.
    r : float, optional
        The warping window in [0, 1] as a fraction of `max(x_timestep, y_timestep)`.

    Returns
    -------
    float
        The dynamic time warping distance.

    See Also
    --------
    dtw_alignment : compute the dtw alignment matrix
    """
    x = check_array(x, ravel_1d=True, ensure_2d=False)
    y = check_array(y, ravel_1d=True, ensure_2d=False)
    return pairwise_distance(x, y, metric="dtw", metric_params={"r": r})


def ddtw_distance(x, y, *, r=1.0):
    """
    Compute the derivative dynamic time warping distance.

    Parameters
    ----------
    x : array-like of shape (x_timestep, )
        The first time series.
    y : array-like of shape (y_timestep, )
        The second time series.
    r : float, optional
        The warping window in [0, 1] as a fraction of max(x_timestep, y_timestep).

    Returns
    -------
    float
        The dynamic time warping distance.

    See Also
    --------
    dtw_distance : compute the dtw distance
    """
    x = check_array(x, ravel_1d=True, ensure_2d=False)
    y = check_array(y, ravel_1d=True, ensure_2d=False)
    return pairwise_distance(x, y, metric="ddtw", metric_params={"r": r})


def wdtw_distance(x, y, *, r=1.0, g=0.05):
    """
    Compute the weighted dynamic time warping distance.

    Parameters
    ----------
    x : array-like of shape (x_timestep, )
        The first time series.
    y : array-like of shape (y_timestep, )
        The second time series.
    r : float, optional
        The warping window in [0, 1] as a fraction of `max(x_timestep, y_timestep)`.
    g : float, optional
        Penalization for points deviating the diagonal.

    Returns
    -------
    float
        The dynamic time warping distance.

    See Also
    --------
    dtw_distance : compute the dtw distance
    """
    x = check_array(x, ravel_1d=True, ensure_2d=False)
    y = check_array(y, ravel_1d=True, ensure_2d=False)
    return pairwise_distance(x, y, metric="wdtw", metric_params={"r": r, "g": g})


def wddtw_distance(x, y, *, r=1.0, g=0.05):
    """
    Compute the weighted derivative dynamic time warping distance.

    Parameters
    ----------
    x : array-like of shape (x_timestep, )
        The first time series.
    y : array-like of shape (y_timestep, )
        The second time series.
    r : float, optional
        The warping window in [0, 1] as a fraction of `max(x_timestep, y_timestep)`.
    g : float, optional
        Penalization for points deviating the diagonal.

    Returns
    -------
    float
        The dynamic time warping distance.

    See Also
    --------
    dtw_distance : compute the dtw distance
    """
    x = check_array(x, ravel_1d=True, ensure_2d=False)
    y = check_array(y, ravel_1d=True, ensure_2d=False)
    return pairwise_distance(x, y, metric="wddtw", metric_params={"r": r, "g": g})


def dtw_envelop(x, *, r=1.0):
    """
    Compute the envelop for LB_keogh.

    Parameters
    ----------
    x : array-like of shape (x_timestep,)
        The time series.
    r : float, optional
        The warping window in [0, 1] as a fraction of `max(x_timestep, y_timestep)`.

    Returns
    -------
    lower : ndarray of shape (x_timestep,)
        The min value of the envelop.
    upper : ndarray of shape (x_timestep,)
        The max value of the envelop.

    References
    ----------
    Keogh, E. (2002).
        Exact indexing of dynamic time warping.
        In 28th International Conference on Very Large Data Bases.
    """
    x = check_array(x, ravel_1d=True, ensure_2d=False, dtype=float, input_name="x")
    warp_size = _compute_warp_size(x.shape[0], r)
    return _dtw_envelop(x, warp_size)


def dtw_lb_keogh(x, y=None, *, lower=None, upper=None, r=1.0):
    """
    LB_keogh lower bound.

    Parameters
    ----------
    x : array-like of shape (x_timestep,)
        The first time series.
    y : array-like of shape (x_timestep,), optional
        The second time series (same size as x).
    lower : ndarray of shape (x_timestep,), optional
        The min value of the envelop.
    upper : ndarray of shape (x_timestep,), optional
        The max value of the envelop.
    r : float, optional
        The warping window in [0, 1] as a fraction of `max(x_timestep, y_timestep)`.

    Returns
    -------
    min_dist : float
        The cumulative minimum distance.
    lb_keogh : ndarray of shape (x_timestep,),
        The lower bound at each time step.

    Notes
    -----
    - if y=None, both lower and upper must be given
    - if y is given, lower and upper are ignored
    - if lower and upper is given and y=None, r is ignored

    References
    ----------
    Keogh, E. (2002).
        Exact indexing of dynamic time warping.
        In 28th International Conference on Very Large Data Bases.
    """
    check_args = dict(ravel_1d=True, ensure_2d=False, dtype=float)
    x = check_array(x, **check_args, input_name="x")
    if y is not None:
        y = check_array(y, **check_args, input_name="y")
        if y.shape[0] != x.shape[0]:
            raise ValueError(
                "x (%d) and y (%d) must have the same number of timesteps"
                % (x.shape[0], y.shape[0])
            )
        lower, upper = dtw_envelop(y, r=r)
    elif lower is None or upper is None:
        raise ValueError("both y, lower and upper can't be None")

    lower = check_array(lower, **check_args, input_name="lower")
    upper = check_array(upper, **check_args, input_name="upper")
    if lower.shape[0] != upper.shape[0] or lower.shape[0] != x.shape[0]:
        raise ValueError(
            "lower (%d), upper (%d) and x (%d) have the same number of timesteps"
            % (lower.shape[0], upper.shape[0], x.shape[0])
        )
    warp_size = _compute_warp_size(x.shape[0], r)
    return _dtw_lb_keogh(x, lower, upper, warp_size)


def dtw_alignment(x, y, *, r=1.0, weight=None, out=None):
    """
    Compute the Dynamic time warping alignment matrix.

    Parameters
    ----------
    x : array-like of shape (x_timestep,)
        The first time series.
    y : array-like of shape (y_timestep,)
        The second time series.
    r : float, optional
        The warping window in [0, 1] as a fraction of max(x_timestep, y_timestep).
    weight : array-like of shape (max(x_timestep, y_timestep), ), optional
        A weighting vector to penalize warping.
    out : array-like of shape (x_timestep, y_timestep), optional
        Store the warping path in this array.

    Returns
    -------
    ndarray of shape (x_timestep, y_timestep)
        The dynamic time warping alignment matrix.

    See Also
    --------
    dtw_distance : compute the dtw distance

    Notes
    -----
    If only the distance between two series is required use `dtw_distance`
    instead.
    """
    x = check_array(x, ravel_1d=True, ensure_2d=False, dtype=float, input_name="x")
    y = check_array(y, ravel_1d=True, ensure_2d=False, dtype=float, input_name="y")
    warp_size = _compute_warp_size(x.shape[0], r, y_size=y.shape[0])
    if weight is not None:
        weight = check_array(
            weight, ravel_1d=True, ensure_2d=False, dtype=float, input_name="weight"
        )
        if weight.shape[0] != max(x.shape[0], y.shape[0]):
            raise ValueError(
                "weight must have the same size as max(x.size, y.size) %d, got %d"
                % (max(x.shape[0], y.shape[0]), weight.shape[0])
            )

    return _dtw_alignment(x, y, warp_size, weight, out)


def wdtw_alignment(x, y, *, r=1.0, g=0.5, out=None):
    """
    Weighted dynamic time warping alignment.

    Parameters
    ----------
    x : array-like of shape (x_timestep,)
        The first time series.
    y : array-like of shape (y_timestep,)
        The second time series.
    r : float, optional
        The warping window in [0, 1] as a fraction of `max(x_timestep, y_timestep)`.
    g : float, optional
        Weighting described by Jeong et. al. (2011)[1]_ using `g` as penalty
        control.

        ::

            w = 1.0 / 1.0 + np.exp(-g * (x - n_samples / 2))
    out : array-like of shape (x_timestep, y_timestep), optional
        Store the warping path in this array.

    Returns
    -------
    ndarray of shape (x_timestep, y_timestep)
        The dynamic time warping alignment matrix.

    References
    ----------
    .. [1] Jeong, Y., Jeong, M., Omitaomu, O. (2021)
        Weighted dynamic time warping for time series classification.
        Pattern Recognition 44, 2231-2240
    """
    weight = jeong_weight(max(x.shape[0], y.shape[0]), g)
    return dtw_alignment(x, y, r=r, weight=weight, out=out)


def jeong_weight(n, g=0.05):
    """
    Weighting described by Jeong et. al. (2011)[1]_.

    Uses `g` as the penalty control.

    ::

        w = 1.0 / 1.0 + np.exp(-g * (x - n_samples / 2))

    Parameters
    ----------
    n : int
        The number of weights.
    g : float, optional
        Penalty control.

    Returns
    -------
    ndarray of shape (n, )
        The weights.

    References
    ----------
    .. [1] Jeong, Y., Jeong, M., Omitaomu, O. (2021)
        Weighted dynamic time warping for time series classification.
        Pattern Recognition 44, 2231-2240
    """
    weight = 1.0 / (1.0 + np.exp(-g * (np.arange(n, dtype=float) - n / 2.0)))
    return weight


def dtw_mapping(x=None, y=None, *, alignment=None, r=1, return_index=False):
    """
    Optimal warping path between two series or from a given alignment matrix.

    Parameters
    ----------
    x : array-like of shape (x_timestep,), optional
        The first time series.
    y : array-like of shape (y_timestep,), optional
        The second time series.
    alignment : ndarray of shape (x_timestep, y_timestep), optional
        Precomputed alignment.
    r : float, optional
        The warping window in [0, 1] as a fraction of `max(x_timestep, y_timestep)`.
    return_index : bool, optional
        Return the indices of the warping path.

    Returns
    -------
    indicator : ndarray of shape (x_timestep, y_timestep)
        Boolean array with the dtw path.
    (x_indices, y_indices) : tuple, optional
        The indices of the first and second dimension of
        the optimal alignment path..

    Notes
    -----
    - either x and y or alignment must be provided
    - if alignment is given x and y are ignored
    - if alignment is given r is ignored
    """
    if alignment is None:
        if x is None or y is None:
            raise ValueError("if alignment=None, neither x or y can be None")
        alignment = dtw_alignment(x, y, r=r)
    else:
        alignment = check_array(
            alignment,
            force_all_finite=False,
            order=None,
            allow_eos=True,  # FIXME!
            input_name="alignment",
        )

    indicator = np.zeros(alignment.shape, dtype=bool)
    i = alignment.shape[0] - 1
    j = alignment.shape[1] - 1
    while i > 0 or j > 0:
        indicator[i, j] = True
        option_diag = alignment[i - 1, j - 1] if i > 0 and j > 0 else np.inf
        option_up = alignment[i - 1, j] if i > 0 else np.inf
        option_left = alignment[i, j - 1] if j > 0 else np.inf
        move = np.argmin([option_diag, option_up, option_left])
        if move == 0:
            i -= 1
            j -= 1
        elif move == 1:
            i -= 1
        else:
            j -= 1
    indicator[0, 0] = True
    if return_index:
        return indicator, indicator.nonzero()
    else:
        return indicator


def dtw_average(
    X,
    *,
    r=1.0,
    g=None,
    sample_weight=None,
    init="random",
    method="mm",
    max_stable=5,
    learning_rate=0.1,
    decay=0.9,
    tol=1e-5,
    max_epoch=50,
    return_cost=False,
    verbose=False,
    random_state=None,
):
    """
    Compute the DTW barycenter average (DBA).

    Parameters
    ----------
    X : array-like of shape (n_samples, n_timestep)
        The samples to average.
    r : float, optional
        The warping window as a fraction of n_timestep.
    g : float, optional
        If set, use the weighted DTW alignment with `g` as penalty control.

        ::

            w(x)= 1.0 / 1.0 + np.exp(-g * (x - m / 2)),
    sample_weight : array-like of shape (n_samples, ), optional
        The sample weight. This influences how much each sample
        contributes to the average.
    init : "random" or array-like of shape (m_timestep, ), optional
        The initial sample used for average.
    method : {"mm", "ssg"}, optional
        The method for computing the DBA.

        - if "mm", use the majorize-minimize mean algorithm [1], which is equivalent to
          the DBA method in [2].

        - if "ssg", use the stochastic subgradient mean algorithm [1].
    max_stable : int, optional
        The maximum number of epoch where the average with lowest cost is unchanged
        if `method='ssg'`.
    learning_rate : float, optional
        The learning rate, if `method="ssg"`.
    decay : float, optional
        The learning rate decay, if `method="ssg"`.
    tol : float, optional
        The minmum change in cost between two epochs, if `method="mm"`.
    max_epoch : int, optional
        The maximum number of epochs.
    return_cost : bool, optional
        Return the averaging cost if `True`.
    verbose : bool, optional
        If set, show runtime information.
    random_state : int or RandomState, optional
        Pseudo-random number generator.

    Returns
    -------
    mean : array-like of shape (m_timestep, ) or (n_timestep, )
        The mean time series.
    cost : float, optional
        Return the cost of the average, if `return_cost=True`.

    Examples
    --------
    >>> from wildboar.datasets import load_two_lead_ecg
    >>> from wildboar.distance.dtw import dtw_average
    >>> X, y = load_two_lead_ecg()
    >>> dtw_average(X[:5], method="mm", random_state=1)
    array([-2.27442791e-01,  3.19807473e-02,  1.77490053e-01,  1.60441308e-01,
            2.31930140e-01,  2.17437783e-01,  2.43925941e-01,  2.60983434e-01,
            2.72118437e-01,  7.73352049e-02, -1.56701557e-02, -5.53269314e-02,
           -7.33366128e-02, -1.09010828e-01, -1.97539989e-01, -1.71443248e-01,
           -1.71443248e-01, -1.71443248e-01, -2.42492836e-01, -1.71408958e-01,
           -1.71408958e-01, -1.71408958e-01, -1.71408958e-01, -1.71408958e-01,
           -1.71408958e-01, -1.71408958e-01, -1.82518334e-01, -3.35671953e-01,
            1.26442901e-01, -7.38342948e-02, -9.11248815e-01, -1.99355168e+00,
           -2.08588712e+00, -2.35954194e+00, -2.78345146e+00, -2.41023092e+00,
           -1.99915956e+00, -1.82717462e+00, -1.82717462e+00, -1.71687181e+00,
           -1.55819192e+00, -1.28805337e+00, -1.06653283e+00, -7.25159669e-01,
           -4.02389872e-01, -2.39410523e-01,  2.34687887e-03,  2.98654485e-01,
            4.85832342e-01,  6.56436416e-01,  7.25302660e-01,  7.77697444e-01,
            8.24606299e-01,  8.76357782e-01,  9.27083874e-01,  9.44590342e-01,
            9.44590342e-01,  9.44590342e-01,  9.44590342e-01,  9.44590342e-01,
            9.64184026e-01,  1.03608265e+00,  1.13964118e+00,  1.33595675e+00,
            1.09954847e+00,  9.61924171e-01,  9.61924171e-01,  9.61924171e-01,
            9.61924171e-01,  9.61924171e-01,  9.61924171e-01,  9.47433305e-01,
            8.29583168e-01,  7.00425122e-01,  5.80524683e-01,  4.70210329e-01,
            4.40259039e-01,  3.59657389e-01,  3.52170730e-01,  3.54666287e-01,
            1.93690730e-01,  2.23968406e-01])
    """
    X = check_array(X, ensure_min_samples=2)
    r = check_scalar(r, "r", numbers.Real, min_val=0.0, max_val=1.0)
    random_state = check_random_state(random_state)

    if isinstance(init, str) and init == "random":
        mean = X[random_state.randint(X.shape[0])].copy()
    elif _is_arraylike(init):
        mean = np.array(init, copy=True)
    else:
        raise ValueError(
            "init must be array-like or 'random', not %r" % type(init).__qualname__
        )

    if sample_weight is not None:
        sample_weight = _check_sample_weight(sample_weight, X)

    if g is None:
        distancefn = partial(pairwise_distance, metric="dtw", metric_params=dict(r=r))
        alignmentfn = partial(dtw_alignment, r=r)
    else:
        g = check_scalar(g, "g", numbers.Real, min_val=0, include_boundaries="neither")
        distancefn = partial(
            pairwise_distance, metric="wdtw", metric_params=dict(r=r, g=g)
        )
        alignmentfn = partial(wdtw_alignment, r=r, g=g)

    def costfn(mean, X):
        cost = distancefn(mean, X)
        if sample_weight is None:
            return np.mean(cost)
        else:
            return np.average(cost, weights=sample_weight)

    if method == "mm":
        mean, cost = _mm_dtw_average(
            mean,
            X,
            sample_weight=sample_weight,
            max_epoch=max_epoch,
            costfn=costfn,
            alignmentfn=alignmentfn,
            tol=tol,
            verbose=verbose,
        )
    elif method == "ssg":
        mean, cost = _ssg_dtw_average(
            mean,
            X,
            sample_weight=sample_weight,
            max_epoch=max_epoch,
            max_stable=max_stable,
            learning_rate=learning_rate,
            decay=decay,
            costfn=costfn,
            alignmentfn=alignmentfn,
            verbose=verbose,
            random_state=random_state,
        )
    else:
        raise ValueError("method must be 'mm' or 'ssg', got %r" % method)

    if return_cost:
        return mean, cost
    else:
        return mean


# Implementation based on Stochastic Subgradient mean algorithm by:
# David Schultz and Brijnesh J. Jain. (2016).
#   Sample Mean Algorithms for Averaging in Dynamic Time Warping Spaces.
#   Zenodo. https://doi.org/10.5281/zenodo.216233
def _ssg_dtw_average(
    mean,
    X,
    *,
    sample_weight,
    max_epoch,
    learning_rate,
    decay,
    max_stable,
    costfn,
    alignmentfn,
    verbose,
    random_state,
):
    best_mean = None
    order = np.arange(X.shape[0])
    min_cost = np.inf
    n_stable = 0

    z = np.empty(mean.shape[0], dtype=float)
    for epoch in range(max_epoch):
        if n_stable > max_stable:
            if verbose:
                print(f"Completed at epoch={epoch} with cost={min_cost}.")
            break

        random_state.shuffle(order)
        for i, o in enumerate(order):
            z.fill(0)
            align_m, align_x = dtw_mapping(alignment=alignmentfn(mean, X[o])).nonzero()

            w = 1.0
            if sample_weight is not None:
                w = sample_weight[o]

            for m, x in zip(align_m, align_x):
                z[m] += mean[m] - X[o, x] * w

            mean -= learning_rate * z

            if epoch == 0:
                learning_rate = decay**i * learning_rate

        cost = costfn(mean, X)
        if cost < min_cost:
            if verbose:
                print(
                    f"New min cost={cost} at epoch={epoch} with "
                    f"learning_rate={learning_rate}."
                )

            min_cost = cost
            n_stable = 0
            best_mean = mean.copy()
        else:
            n_stable += 1

    return best_mean, min_cost


# Implementation based on Majorize-Minimize mean algorithm by:
# David Schultz and Brijnesh J. Jain. (2016).
#   Sample Mean Algorithms for Averaging in Dynamic Time Warping Spaces.
#   Zenodo. https://doi.org/10.5281/zenodo.216233
def _mm_dtw_average(
    mean, X, *, sample_weight, max_epoch, costfn, alignmentfn, tol, verbose
):
    prev_cost = np.inf
    cost = costfn(mean, X)
    z = np.empty(mean.shape[0], dtype=float)
    V = np.empty(mean.shape[0], dtype=float)
    for epoch in range(max_epoch):
        z.fill(0)
        V.fill(0)
        for i in range(X.shape[0]):
            align_m, align_x = dtw_mapping(alignment=alignmentfn(mean, X[i])).nonzero()
            w = 1.0
            if sample_weight is not None:
                w = sample_weight[i]
            for m, x in zip(align_m, align_x):
                V[m] += w
                z[m] += X[i, x] * w

        mean = z / V

        prev_cost = cost
        cost = costfn(mean, X)

        if abs(prev_cost - cost) < tol:
            if verbose:
                print(f"Complete at epoch={epoch} with cost={cost}.")
            break

    return mean, cost



# cython: language_level=3

# Authors: Isak Samsten
# License: BSD 3 clause

cdef struct IncStats:
    double n_samples
    double mean
    double sum_square
    double sum

cdef void inc_stats_init(IncStats *inc_stats) noexcept nogil

cdef void inc_stats_add(IncStats *inc_stats, double weight, double value) noexcept nogil

cdef void inc_stats_remove(IncStats *inc_stats, double weight, double value) noexcept nogil

cdef double inc_stats_mean(IncStats *inc_stats) noexcept nogil

cdef double inc_stats_variance(IncStats *inc_stats, bint sample=*) noexcept nogil

cdef double inc_stats_n_samples(IncStats *inc_stats) noexcept nogil

cdef double inc_stats_sum(IncStats *inc_stats) noexcept nogil

cdef void cumulative_mean_std(
    const double *x,
    Py_ssize_t x_length, 
    Py_ssize_t y_length, 
    double *x_mean, 
    double *x_std
) noexcept nogil

cdef double find_min(const double *x, Py_ssize_t n, Py_ssize_t *min_index=*) noexcept nogil

cdef double mean(const double *x, Py_ssize_t length) noexcept nogil

cdef double variance(const double *x, Py_ssize_t length) noexcept nogil

cdef double slope(const double *x, Py_ssize_t length) noexcept nogil

cdef double covariance(const double *x, const double *y, Py_ssize_t length) noexcept nogil

cdef void auto_correlation(const double *x, Py_ssize_t n, double *out) noexcept nogil

cdef void _auto_correlation(const double *x, Py_ssize_t n, double *out, complex *fft) noexcept nogil

cdef Py_ssize_t next_power_of_2(Py_ssize_t n) noexcept nogil

cdef int welch(
    const double *x, 
    Py_ssize_t size, 
    int NFFT, 
    double Fs, 
    double *window, 
    int windowWidth,
    double *Pxx, 
    double *f,
) noexcept nogil


cdef void fast_mean_std(
    const double* data,
    Py_ssize_t length,
    double *mean,
    double* std,
) noexcept nogil



# Authors: Isak Samsten
# License: BSD 3 clause

import abc
import math
import numbers
import warnings
from copy import deepcopy
from functools import partial

import numpy as np
from sklearn.neighbors import NearestNeighbors
from sklearn.utils._param_validation import Interval, StrOptions
from sklearn.utils.validation import check_is_fitted, check_random_state

from ...base import BaseEstimator, CounterfactualMixin, ExplainerMixin
from ...distance import pairwise_subsequence_distance
from ...distance.dtw import (
    dtw_alignment,
    dtw_distance,
    dtw_mapping,
    wdtw_alignment,
    wdtw_distance,
)


class TargetEvaluator(abc.ABC):
    """
    Evaluate if a sample is a counterfactual.

    Parameters
    ----------
    estimator : object
        The estimator.
    """

    def __init__(self, estimator):
        self.estimator = estimator

    def is_counterfactual(self, x, y):
        """
        Return true if x is a counterfactual of label y.

        Parameters
        ----------
        x : ndarray of shape (n_timestep,)
            The counterfactual sample.
        y : object
            The counterfactual label.

        Returns
        -------
        bool
            Return true if counterfactual valid.
        """
        return self._is_counterfactual(x.reshape(1, -1), y)

    @abc.abstractmethod
    def _is_counterfactual(self, x, y):
        pass


class PredictEvaluator(TargetEvaluator):
    """Evaluate if a counterfactual is predicted as y."""

    def _is_counterfactual(self, x, y):
        return self.estimator.predict(x)[0] == y


class ProbabilityEvaluator(TargetEvaluator):
    """
    Evaluate the probability threshold.

    Parameters
    ----------
    estimator : object
        The estimator.
    probability : float, optional
        The minimum probability of the predicted label.
    """

    def __init__(self, estimator, probability=0.5):
        super().__init__(estimator)
        self.probability = probability

    def _is_counterfactual(self, x, y):
        if not hasattr(self.estimator, "predict_proba"):
            raise ValueError("estimator must support predict_proba")

        y_pred = self.estimator.predict_proba(x)
        y_idx = (self.estimator.classes_ == y).nonzero()[0][0]
        y_prob = y_pred[0, y_idx]
        return y_prob > self.probability


class PrototypeSampler(abc.ABC):
    """
    Sample and refine counterfactuals.

    Parameters
    ----------
    x : ndarray of shape (n_samples, n_timestep)
        The data samples labeled as y.
    y : object
        The label of the samples in x.
    """

    def __init__(self, x, y, prototype_indices, metric_transform):
        self.x = x
        self.y = y
        self.metric_transform = metric_transform
        self.prototype_indices = prototype_indices

    def _get_random_index(self, random_state):
        """
        Return a random index in the initial prototype sample.

        Parameters
        ----------
        random_state : RandomState
            The random state.

        Returns
        -------
        int
            The index.
        """
        return self.prototype_indices[
            random_state.randint(self.prototype_indices.shape[0])
        ]

    @abc.abstractmethod
    def sample(self, o, random_state):
        """
        Sample an example.

        Parameters
        ----------
        o : ndarray of shape (n_timestep,)
            The current counterfactual sample.
        random_state : RandomState
            The random state.

        Returns
        -------
        ndarray of shape (n_timestep,)
            A prototype of the counterfactual label.
        """
        pass

    def move(self, o, p):
        """
        Move the current counterfactual toward the prototype.

        Parameters
        ----------
        o : ndarray of shape (n_timestep,)
            The current counterfactual sample.
        p : ndarray of shape (n_timestep,)
            The prototype of the counterfactual label.

        Returns
        -------
        ndarray of shape (n_timestep,)
            The new counterfactual moved towards the prototype.
        """
        return self.metric_transform.move(o, p)

    def sample_move(self, o, random_state):
        """
        Sampla a prototype and move the counterfactual towards the prototype.

        Parameters
        ----------
        o : ndarray of shape (n_timestep,)
            The current counterfactual sample.
        random_state : RandomState
            The random state.

        Returns
        -------
        ndarray of shape (n_timestep,)
            The new counterfactual moved towards the prototype.
        """
        p = self.sample(o, random_state)
        return self.move(o, p)


class UniformPrototypeSampler(PrototypeSampler):
    """Sample a prototype uniformly at random from the initial prototype sample."""

    def sample(self, _o, random_state):
        return self.x[self._get_random_index(random_state)]


class KNearestPrototypeSampler(PrototypeSampler):
    """Sample a prototype among the samples closest to the current counterfactual."""

    def __init__(self, x, y, prototype_indicies, metric_transform):
        super().__init__(x, y, prototype_indicies, metric_transform)
        n_prototypes = prototype_indicies.size
        if isinstance(self.metric_transform, EuclideanTransform):
            self.nearest_neighbors = NearestNeighbors(
                metric="euclidean", n_neighbors=n_prototypes
            )
        elif isinstance(self.metric_transform, DynamicTimeWarpTransform):
            dtw = partial(dtw_distance, r=self.metric_transform.r)
            self.nearest_neighbors = NearestNeighbors(
                metric=dtw, n_neighbors=n_prototypes
            )
        elif isinstance(self.metric_transform, WeightedDynamicTimeWarpTransform):
            wdtw = partial(
                wdtw_distance, r=self.metric_transform.r, g=self.metric_transform.g
            )
            self.nearest_neighbors = NearestNeighbors(
                metric=wdtw, n_neighbors=n_prototypes
            )
        else:
            raise ValueError("unsupported metric")

        self.nearest_neighbors.fit(x)

    def nearest_index(self, o, random_state):
        """Return the index of the closest sample.

        Parameters
        ----------
        o : ndarray of shape (n_timestep,)
            The current counterfactual sample

        Returns
        -------
        int : an index
        """
        nearest = self.nearest_neighbors.kneighbors(
            o.reshape(1, -1), return_distance=False
        ).reshape(-1)
        return nearest[random_state.randint(nearest.shape[0])]

    def sample(self, o, random_state):
        return self.x[self.nearest_index(o, random_state)]


class ShapeletPrototypeSampler(PrototypeSampler):
    """
    Sample shapelet prototypes.

    Parameters
    ----------
    x : ndarray of shape (n_samples, n_timestep)
        The data samples
    y : object
        The label of the samples in x
    min_shapelet_size : float
        The minimum shapelet size
    max_shapelet_size : float
            The maximum shapelet size
    """

    def __init__(
        self,
        x,
        y,
        prototype_indicies,
        metric_transform,
        min_shapelet_size=0,
        max_shapelet_size=1,
    ):
        super().__init__(x, y, prototype_indicies, metric_transform)
        self.min_shapelet_size = min_shapelet_size
        self.max_shapelet_size = max_shapelet_size

    def sample_shapelet(self, p, random_state):
        """
        Sample a shapelet from x.

        Parameters
        ----------
        p : ndarray of shape (n_timestep,)
             The prototype sample

        Returns
        -------
        shapelet : ndarray
            A shapelet
        """
        n_timestep = self.x.shape[-1]
        min_shapelet_size = max(
            min(2, n_timestep), math.ceil(n_timestep * self.min_shapelet_size)
        )
        max_shapelet_size = math.ceil(n_timestep * self.max_shapelet_size)
        shapelet_length = random_state.randint(min_shapelet_size, max_shapelet_size)
        start_index = random_state.randint(0, n_timestep - shapelet_length)
        return p[start_index : (start_index + shapelet_length)]

    def sample(self, _o, random_state):
        return self.sample_shapelet(
            self.x[self._get_random_index(random_state)], random_state
        )

    def move(self, o, p):
        """Move the best matching shapelet towards the shapelet prototype.

        Parameters
        ----------
        o : ndarray of shape (n_timestep,)
            The counterfactual sample
        p : ndarray
            The prototype shapelet

        Returns
        -------
        new_counterfactual : ndarray of shape (n_timestep,)
            The new counterfactual moved towards the prototype
        """
        if isinstance(self.metric_transform, EuclideanTransform):
            metric = "euclidean"
            metric_params = {}
        elif isinstance(self.metric_transform, DynamicTimeWarpTransform):
            metric = "dtw"
            metric_params = {"r": self.metric_transform.r}
        elif isinstance(self.metric_transform, WeightedDynamicTimeWarpTransform):
            metric = "wdtw"
            metric_params = {"r": self.metric_transform.r, "g": self.metric_transform.g}
        else:
            raise ValueError("unsupported metric")

        # Find the best matching position in
        min_dist, best_match = pairwise_subsequence_distance(
            p,
            o,
            metric=metric,
            metric_params=metric_params,
            return_index=True,
        )
        o[best_match : best_match + p.shape[0]] = self.metric_transform.move(
            o[best_match : best_match + p.shape[0]], p
        )

        return o


class KNearestShapeletPrototypeSampler(PrototypeSampler):
    """Combines the KNearestPrototypeSample and the ShapeletPrototypeSampler.

    The prototype samples are sampled among the nearest neighbors of the
    counterfactual.
    """

    def __init__(
        self,
        x,
        y,
        prototype_indicies,
        metric_transform,
        min_shapelet_size=0,
        max_shapelet_size=1,
    ):
        super().__init__(x, y, prototype_indicies, metric_transform)
        self.nearest_sampler = KNearestPrototypeSampler(
            x, y, prototype_indicies, metric_transform
        )
        self.shapelet_sampler = ShapeletPrototypeSampler(
            x,
            y,
            prototype_indicies,
            metric_transform,
            min_shapelet_size,
            max_shapelet_size,
        )

    def sample(self, o, random_state):
        p = self.nearest_sampler.nearest_index(o, random_state)
        return self.shapelet_sampler.sample_shapelet(self.x[p], random_state)

    def move(self, o, p):
        return self.shapelet_sampler.move(o, p)


class MetricTransform(abc.ABC):
    """Move a time series towards a prototype."""

    def __init__(self, gamma):
        """Construct a new transformer.

        Parameters
        ----------
        gamma : float
            The strength of the move, where values close to 0 means that the
            sample is moved less and values closer to 1 mean that the sample
            is moved more.
        """
        self.gamma = gamma

    @abc.abstractmethod
    def move(self, o, p):
        """Move the sample o towards p.

        Parameters
        ----------
        o : ndarray of shape (n_timestep,)
            An array, the time series to move
        p : ndarray of shape (n_timestep,)
            An array, the time series to move towards

        Returns
        -------
        ndarray : an array, the result of moving o closer to p
        """
        pass


class EuclideanTransform(MetricTransform):
    """Transform a sample by moving it closer in euclidean space."""

    def move(self, o, p):
        return ((1 - self.gamma) * o) + self.gamma * p


class DynamicTimeWarpTransform(MetricTransform):
    """Transform a sample by moving it closer using the dtw alignment matrix."""

    def __init__(self, gamma, r=1.0):
        super().__init__(gamma)
        self.r = r

    def move(self, o, p):
        _, indices = dtw_mapping(alignment=self._get_alignment(o, p), return_index=True)

        # This a fast but somewhat crude approximation of DBA (as implemented) in
        # dtw_average with sample_weight=[(1 - gamma), gamma] and init=0 but
        # significantly faster to compute.
        result = o * (1 - self.gamma)
        weight = np.ones(o.shape[0]) * (1 - self.gamma)
        for i, j in zip(indices[0], indices[1]):
            result[i] += self.gamma * p[j]
            weight[i] += self.gamma

        return result / weight

    def _get_alignment(self, o, p):
        return dtw_alignment(o, p, r=self.r)


class WeightedDynamicTimeWarpTransform(DynamicTimeWarpTransform):
    def __init__(self, gamma, r=1, g=0.05):
        super().__init__(gamma, r)
        self.g = g

    def _get_alignment(self, o, p):
        return wdtw_alignment(o, p, r=self.r, g=self.g)


_METRIC_TRANSFORM = {
    "euclidean": EuclideanTransform,
    "dtw": DynamicTimeWarpTransform,
    "wdtw": WeightedDynamicTimeWarpTransform,
}

_PROTOTYPE_SAMPLER = {
    "sample": UniformPrototypeSampler,
    "nearest": KNearestPrototypeSampler,
    "shapelet": ShapeletPrototypeSampler,
    "nearest_shapelet": KNearestShapeletPrototypeSampler,
}


class PrototypeCounterfactual(CounterfactualMixin, ExplainerMixin, BaseEstimator):
    """Model agnostic approach for constructing counterfactual explanations.

    Attributes
    ----------
    estimator_ : object
        The estimator for which counterfactuals are computed
    classes_ : ndarray
        The classes
    partitions_ : dict
        Dictionary of classes and PrototypeSampler
    target_ : TargetEvaluator
        The target evaluator

    References
    ----------
    Samsten, Isak (2020).
        Model agnostic time series counterfactuals
    """

    _parameter_constraints: dict = {
        "metric": [StrOptions(_METRIC_TRANSFORM.keys())],
        "r": [None, Interval(numbers.Real, 0, 1.0, closed="both")],
        "g": [None, Interval(numbers.Real, 0, None, closed="right")],
        "max_iter": [Interval(numbers.Integral, 1, None, closed="left")],
        "step_size": [Interval(numbers.Real, 0, 1, closed="right")],
        "n_prototypes": [
            StrOptions({"auto"}),
            Interval(numbers.Integral, 1, None, closed="left"),
            Interval(numbers.Real, 0, 1, closed="right"),
        ],
        "target": [
            StrOptions({"predict", "auto"}, deprecated={"auto"}),
            Interval(numbers.Real, 0.5, 1, closed="right"),
        ],
        "method": [StrOptions(_PROTOTYPE_SAMPLER.keys())],
        "min_shapelet_size": [Interval(numbers.Real, 0, 1, closed="left")],
        "max_shapelet_size": [Interval(numbers.Real, 0, 1, closed="right")],
        "random_state": ["random_state"],
        "verbose": ["verbose"],
    }

    def __init__(
        self,
        metric="euclidean",
        *,
        r=1.0,
        g=0.05,
        max_iter=100,
        step_size=0.1,
        n_prototypes="auto",
        target="predict",
        method="sample",
        min_shapelet_size=0.0,
        max_shapelet_size=1.0,
        random_state=None,
        verbose=False,
    ):
        """Crate a new model agnostic counterfactual explainer.

        Parameters
        ----------
        metric : {'euclidean', 'dtw', 'wdtw'}, optional
            The metric used to move the samples
        r : float, optional
            The warping window size, if metric='dtw' or metric='wdtw'
        g : float, optional
            Penalty control for weighted DTW, if metric='wdtw'
        max_iter : int, optional
            The maximum number of iterations
        step_size : float, optional
            The step size when moving samples toward class prototypes
        n_prototypes : int, float or str, optional
            The number of initial prototypes to sample from
        target : float or {'predict'}, optional
            The target evaluation of counterfactuals:

            - if 'predict' the counterfactual prediction must return the correct
              label
            - if float, the counterfactual prediction probability must
              exceed target value
        method : {'sample', 'shapelet', 'nearest', 'nearest_shapelet'}, optional
            Method for selecting prototypes

            - if 'sample' a prototype is sampled among the initial prototypes
            - if 'shapelet' a prototype shapelet is sampled among the initial
              prototypes
            - if 'nearest' a prototype is sampled from the closest n prototypes
            - if 'nearest_shapelet' a prototype shapelet is sampled from the
              closest n prototypes
        min_shapelet_size : float, optional
            Minimum shapelet size, if method='shapelet' or 'nearest_shapelet'
        max_shapelet_size : float, optional
            Maximum shapelet size, if method='shapelet' or 'nearest_shapelet'
        random_state : RandomState or int, optional
            Pseudo-random number for consistency between different runs
        """
        self.random_state = random_state
        self.metric = metric
        self.r = r
        self.g = g
        self.max_iter = max_iter
        self.step_size = step_size
        self.n_prototypes = n_prototypes
        self.method = method
        self.min_shapelet_size = min_shapelet_size
        self.max_shapelet_size = max_shapelet_size
        self.target = target
        self.verbose = verbose

    def fit(self, estimator, x, y):
        if x is None or y is None:
            raise ValueError("Both training samples and labels are required.")

        self._validate_params()
        estimator = self._validate_estimator(estimator)
        x, y = self._validate_data(x, y, reset=False, dtype=float)

        metric_params = {}
        if self.metric in ["dtw", "wdtw"]:
            metric_params["r"] = self.r
        if self.metric == ["wdtw"]:
            metric_params["g"] = self.g

        random_state = check_random_state(self.random_state)
        metric_transform = _METRIC_TRANSFORM[self.metric](
            self.step_size,
            **metric_params,
        )
        Sampler = _PROTOTYPE_SAMPLER[self.method]

        self.estimator_ = deepcopy(estimator)
        self.classes_ = np.unique(y)
        self.random_state_ = random_state.randint(np.iinfo(np.int32).max)
        if self.target in {"auto", "predict"}:
            if self.target == "auto":
                warnings.warn(
                    f"The parameter value 'auto' for target of {type(self).__name__} "
                    "has been renamed to 'predict' in 1.2 and will be removed in 1.4.",
                    DeprecationWarning,
                )
            self.target_ = PredictEvaluator(self.estimator_)
        else:
            self.target_ = ProbabilityEvaluator(self.estimator_, self.target)

        if self.method in ["shapelet", "nearest_shapelet"]:
            if self.min_shapelet_size > self.max_shapelet_size:
                raise ValueError(
                    f"The parameter min_shapelet_size of {type(self).__name__} must be "
                    "<= max_shapelet_size."
                )

            method_params = {
                "min_shapelet_size": self.min_shapelet_size,
                "max_shapelet_size": self.max_shapelet_size,
            }
        else:
            method_params = {}

        self.partitions_ = {}
        for c in self.classes_:
            x_partition = x[y == c]
            if self.n_prototypes == "auto":
                n_prototypes = x_partition.shape[0]
            elif isinstance(self.n_prototypes, numbers.Integral):
                n_prototypes = max(1, min(self.n_prototypes, x_partition.shape[0]))
            else:
                n_prototypes = math.ceil(self.n_prototypes * x_partition.shape[0])

            prototype_indicies = np.arange(x_partition.shape[0])
            random_state.shuffle(prototype_indicies)
            prototype_indicies = prototype_indicies[:n_prototypes]
            self.partitions_[c] = Sampler(
                x_partition,
                c,
                prototype_indicies,
                metric_transform,
                **method_params,
            )

    def explain(self, x, y):
        check_is_fitted(self)
        random_state = check_random_state(self.random_state_)
        x, y = self._validate_data(x, y, reset=False, dtype=float)
        counterfactuals = np.empty(x.shape, dtype=x.dtype)
        for i in range(x.shape[0]):
            if self.verbose:
                print(f"Computing counterfactual for the {i}th sample.")

            counterfactuals[i] = self._transform_sample(
                x[i], y[i], random_state.randint(np.iinfo(np.int32).max)
            )

        return counterfactuals

    def _transform_sample(self, x, y, random_state):
        random_state = check_random_state(random_state)
        sampler = self.partitions_[y]
        o = x.copy()
        n_iter = 0
        while not self.target_.is_counterfactual(o, y) and n_iter < self.max_iter:
            if self.verbose and n_iter % (self.max_iter // 10) == 0:
                print(f"Running {n_iter}/{self.max_iter}...")

            o = sampler.sample_move(o, random_state)
            n_iter += 1

        if self.verbose:
            print(f"Completed after {n_iter} iterations.")
            if n_iter == self.max_iter:
                print(
                    "The counterfactual explain reached max_iter, increase step_size "
                    "or max_iter for convergence."
                )
        return o

    def _more_tags():
        return {"requires_y": True}



import math
import numbers

import numpy as np
from sklearn import clone
from sklearn.metrics import accuracy_score
from sklearn.neighbors import LocalOutlierFactor

from ..distance import paired_distance, pairwise_distance
from ..explain._importance import _intervals
from ..explain.counterfactual import proximity
from ..transform._sax import (
    piecewice_aggregate_approximation,
    symbolic_aggregate_approximation,
)
from ..utils.validation import check_array


def _estimate_plausability(estimator, X_counterfactuals, method):
    if method == "score":
        return np.mean(estimator.decision_function(X_counterfactuals))
    elif method == "accuracy":
        y_true = np.broadcast_to(1, X_counterfactuals.shape[0])
        return accuracy_score(y_true, estimator.predict(X_counterfactuals))
    else:
        raise ValueError("method must be 'average', or 'accuracy', " "got %r" % method)


def plausability_score(
    x_plausible,
    x_counterfactuals,
    *,
    y_plausible=None,
    y_counterfactual=None,
    estimator=None,
    method="accuracy",
    average=True,
):
    """
    Compute plausibility score.

    Parameters
    ----------
    x_plausible : array-like of shape (n_samples, n_timesteps)
        The plausible samples, typically the training or testing samples.
    x_counterfactuals : array-like of shape (m_samples, n_timesteps)
        The counterfactual samples.
    y_plausible : array-like of shape (n_samples, ), optional
        The labels of the plausible samples.
    y_counterfactual : array-like of shape (m_samples, ), optional
        The desired label of the counterfactuals.
    estimator : estimator, optional
        The outlier estimator, must implement `fit` and `predict`. If None,
        we use LocalOutlierFactor.

        - if score="mean", the estimator must also implement `decision_function`.
    method : {'score', 'accuracy'}, optional
        The score function.
    average : bool, optional
        If True, return the average score for all labels in y_counterfactual;
        otherwise, return the score for the individual labels (ordered as np.unique).

    Returns
    -------
    ndarray or float
        The plausability.

        - if method='scores', the mean score is returned, with larger score incicating
          better performance.
        - if method='accuracy', the fraction of plausible counterfactuals are returned.
        - if y_counterfactual is None and average=False, the scores or accuracy for each
          counterfactual label is returned.

    References
    ----------
    Delaney, E., Greene, D., & Keane, M. T. (2020).
        Instance-based Counterfactual Explanations for Time Series Classification.
        arXiv, 2009.13211v2.
    """
    if estimator is None:
        estimator = LocalOutlierFactor(
            n_neighbors=math.ceil(np.sqrt(x_plausible.shape[0])), novelty=True
        )
    else:
        estimator = clone(estimator)

    x_plausible = check_array(x_plausible, allow_3d=True)
    x_counterfactuals = check_array(x_counterfactuals, allow_3d=True)
    if x_plausible.shape[-1] != x_counterfactuals.shape[-1]:
        raise ValueError(
            "X_plausible (%s) and X_counterfactuals (%s) must have the same number "
            "of timesteps." % (x_plausible.shape, x_counterfactuals.shape)
        )
    if y_counterfactual is None:
        estimator.fit(x_plausible)
        return _estimate_plausability(estimator, x_counterfactuals, method)
    else:
        if y_plausible is None:
            raise ValueError(
                "if y_counterfactual is given, y_plausible must also be given"
            )

        y_plausible = check_array(
            y_plausible, ensure_2d=False, ravel_1d=True, dtype=None
        )
        y_counterfactual = check_array(
            y_counterfactual, ensure_2d=False, ravel_1d=True, dtype=None
        )

        labels = np.unique(y_counterfactual)
        scores = []
        for label in labels:
            X_plausible_label = x_plausible[y_plausible == label]
            if X_plausible_label.shape[0] == 0:
                raise ValueError(f"Not enough plausable samples with label={label}.")

            label_estimator = clone(estimator)
            label_estimator.fit(X_plausible_label)
            scores.append(
                _estimate_plausability(
                    label_estimator,
                    x_counterfactuals[y_counterfactual == label],
                    method,
                )
            )

        scores = np.array(scores, dtype=float)
        return scores.mean() if average else scores


def relative_proximity_score(
    x_native,
    x_factual,
    x_counterfactual,
    *,
    y_native=None,
    y_counterfactual=None,
    metric="euclidean",
    metric_params=None,
    average=True,
):
    """
    Compute relative proximity score.

    The relative proximity score captures the mean proximity of counterfactual
    and test sample pairs over mean proximity of the closest native
    counterfactual. The lower the score, the better.

    Parameters
    ----------
    x_native : array-like of shape (n_natives, n_timesteps)
        The native counterfactual candidates. If y_counterfactual is None, the full
        array is considered as possible native counterfactuals. Typically, native
        counterfactual candidates correspond to samples which are labeled as the
        desired counterfactual label.
    x_factual : array-like of shape (n_counterfactuals, n_timesteps)
        The factual samples, i.e., the samples for which the counterfactuals
        where computed.
    x_counterfactual : array-like of shape (n_counterfactuals, n_timesteps)
        The counterfactual samples.
    y_native : array-like of shape (n_natives, ), optional
        The label of the native counterfactual candidates.
    y_counterfactual : array-like of shape (n_counterfactuals, ), optional
        The desired counterfactual label.
    metric : str or callable, optional
        The distance metric

        See ``_METRICS.keys()`` for a list of supported metrics.
    metric_params : dict, optional
        Parameters to the metric.

        Read more about the parameters in the
        :ref:`User guide <list_of_metrics>`.
    average : bool, optional
        Average the relative proximity of all labels in y_counterfactual.

    Returns
    -------
    ndarray or float
        The relative proximity. If avarege=False and y_counterfactual is not None,
        return the relative proximity for each counterfactual label.

    Notes
    -----
    The samples in `x_counterfactual` and `x_factual` should be aligned such
    that the i:th counterfacutal sample is derived from the i:th factual sample.

    References
    ----------
    Smyth, B., & Keane, M. T. (2021).
        A Few Good Counterfactuals: Generating Interpretable, Plausible and Diverse
        Counterfactual Explanations. arXiv, 2101.09056v1.
    """
    x_native = check_array(x_native)
    x_factual = check_array(x_factual)
    x_counterfactual = check_array(x_counterfactual)

    cf_dist = paired_distance(
        x_factual, x_counterfactual, metric=metric, metric_params=metric_params
    ).mean()
    if y_counterfactual is None:
        native_dist = pairwise_distance(
            x_native, x_factual, metric=metric, metric_params=metric_params
        )
        return cf_dist / native_dist.min(axis=0).mean()
    else:
        if y_native is None:
            raise ValueError("if y_counterfactual is give, y_native must also be given")

        y_native = check_array(y_native, ensure_2d=False, ravel_1d=True, dtype=None)
        y_counterfactual = check_array(
            y_counterfactual, ensure_2d=False, ravel_1d=True, dtype=None
        )
        cf_labels = np.unique(y_counterfactual)
        native_dists = []
        for label in cf_labels:
            X_native_cf_label = x_native[y_native == label]
            if X_native_cf_label.shape[0] == 0:
                raise ValueError(f"Not enough native samples with label={label}.")

            native_dist = pairwise_distance(
                X_native_cf_label,
                x_factual[y_counterfactual == label],
                metric=metric,
                metric_params=metric_params,
            )
            native_dists.append(native_dist.min(axis=0).mean())

        native_dists = np.array(native_dists, dtype=float)
        return cf_dist / (native_dists.mean() if average else native_dists)


def proximity_score(
    x_factual,
    x_counterfactual,
    metric="normalized_euclidean",
    metric_params=None,
):
    """
    Compute proximity score.

    The closer the counterfactual is to the original, the lower the score.

    Parameters
    ----------
    x_factual : array-like of shape (n_samples, n_timestep)
        The true samples.
    x_counterfactual : array-like of shape (n_samples, n_timestep)
        The counterfactual samples.
    metric : str or callable, optional
        The distance metric

        See ``_METRICS.keys()`` for a list of supported metrics.
    metric_params : dict, optional
        Parameters to the metric.

        Read more about the parameters in the
        :ref:`User guide <list_of_metrics>`.

    Returns
    -------
    float
        The mean proximity.

    Notes
    -----
    The samples in `x_counterfactual` and `x_factual` should be aligned such
    that the i:th counterfacutal sample is derived from the i:th factual sample.

    References
    ----------
    Delaney, E., Greene, D., & Keane, M. T. (2020).
        Instance-based Counterfactual Explanations for Time Series Classification.
        arXiv, 2009.13211v2.
    Karlsson, I., Rebane, J., Papapetrou, P., & Gionis, A. (2020).
        Locally and globally explainable time series tweaking.
        Knowledge and Information Systems, 62(5), 1671-1700.
    """
    x_true = check_array(x_factual, allow_3d=True)
    x_counterfactuals = check_array(x_counterfactual, allow_3d=True)

    return np.mean(
        proximity(x_true, x_counterfactuals, metric=metric, metric_params=metric_params)
    )


def compactness_score(
    x_factual,
    x_counterfactual,
    *,
    window=None,
    n_bins=None,
    atol=1.0e-8,
    average=True,
):
    """
    Compute compactness score.

    The compactness of counterfactuals as measured by the fraction of changed
    timesteps. The fewer timesteps have changed between the original and the
    counterfactual, the lower the score.

    Parameters
    ----------
    x_factual : array-like of shape (n_samples, n_timesteps) \
            or (n_samples, n_dims, n_timeteps)
        The true samples.
    x_counterfactual : array-like of shape (n_samples, n_timesteps) \
            or (n_samples, n_dims, n_timeteps)
        The counterfactual samples.
    window : int, optional
        If set, evaluate the difference between windows of specified size.
    n_bins : int, optional
        If set, evaluate the set overlap of SAX transformed series.
    atol : float, optional
        The absolute tolerance.
    average : bool, optional
        Compute average score over all dimensions.

    Returns
    -------
    float
        The compactness score. Lower score indicates more compact counterfactuals.

    Notes
    -----
    The samples in `x_counterfactual` and `x_factual` should be aligned such
    that the i:th counterfacutal sample is derived from the i:th factual sample.

    References
    ----------
    Karlsson, I., Rebane, J., Papapetrou, P., & Gionis, A. (2020).
        Locally and globally explainable time series tweaking.
        Knowledge and Information Systems, 62(5), 1671-1700.
    """
    x_factual = check_array(x_factual, allow_3d=True)
    x_counterfactual = check_array(x_counterfactual, allow_3d=True)
    if x_factual.shape != x_counterfactual.shape:
        raise ValueError(
            "x_true (%s) and x_counterfactuals (%s) must have the same shape."
            % (x_factual.shape, x_counterfactual.shape)
        )

    if window is not None:
        if n_bins is not None:

            def score(x_counterfactual, x_factual):
                x_counterfactual = symbolic_aggregate_approximation(
                    x_counterfactual, window=window, n_bins=n_bins
                )
                x_factual = symbolic_aggregate_approximation(
                    x_factual, window=window, n_bins=n_bins
                )
                return x_counterfactual == x_factual

        else:

            def score(x_counterfactual, x_factual):
                x_counterfactual = piecewice_aggregate_approximation(
                    x_counterfactual, window=window
                )
                x_factual = piecewice_aggregate_approximation(x_factual, window=window)
                return np.isclose(x_counterfactual, x_factual, rtol=0, atol=atol)

    else:

        def score(x_counterfactual, x_true):
            return np.isclose(x_counterfactual, x_true, rtol=0, atol=atol)

    return 1 - np.mean(score(x_counterfactual, x_factual), axis=None if average else 0)


def validity_score(y_predicted, y_counterfactual, sample_weight=None):
    """
    Compute validity score.

    The number counterfactuals that have the desired label.

    Parameters
    ----------
    y_predicted : array-like of shape (n_samples, )
        The predicted label.
    y_counterfactual : array-like of shape (n_samples, )
        The predicted label.
    sample_weight : array-like of shape (n_samples, ), optional
        The sample weight.

    Returns
    -------
    float
        The fraction of counterfactuals with the correct label. Larger is better.

    References
    ----------
    Delaney, E., Greene, D., & Keane, M. T. (2020).
        Instance-based Counterfactual Explanations for Time Series Classification.
        arXiv, 2009.13211v2.
    Karlsson, I., Rebane, J., Papapetrou, P., & Gionis, A. (2020).
        Locally and globally explainable time series tweaking.
        Knowledge and Information Systems, 62(5), 1671-1700.
    """
    return accuracy_score(y_predicted, y_counterfactual, sample_weight=sample_weight)


def redudancy_score(
    estimator,
    x_factual,
    x_counterfactual,
    y_counterfactual,
    *,
    n_intervals="sqrt",
    window=None,
    average=True,
):
    """
    Compute the redudancy score.

    Redundancy is measure of how much impact non-overlapping intervals has
    in the construction of the counterfactuals.

    Parameters
    ----------
    estimator : Estimator
        The estimator counterfactuals are computed for.
    x_factual : array-like of shape (n_samples, n_timestep)
        The factual samples, i.e., samples for which counterfactuals
        are computed.
    x_counterfactual : array-like of shape (n_samples, n_timestep)
        The counterfactual samples.
    y_counterfactual : array-like of shape (n_samples, )
        The desired counterfactual label.
    n_intervals : {"sqrt", "log2"}, int or float, optional
        The number of intervals.
    window : int, optional
        The size of an interval. If set, `n_intervals` is ignored.
    average : bool, optional
        Return the average redundancy over all intervals.

    Returns
    -------
    ndarray of shape (n_intervals, ) or float
        The redundancy of each interval, expressed as the fraction
        of samples that have the same label if the interval is replaced
        with the corresponding interval of the factual sample. If `average`
        is True, return a single float.

    Notes
    -----
    The samples in `x_counterfactual` and `x_factual` should be aligned such
    that the i:th counterfacutal sample is derived from the i:th factual sample.
    """
    if window is not None:
        if window > x_counterfactual.shape[-1]:
            raise ValueError(
                "The window parameter must be <= X_counterfactual.shape[-1]"
            )
        n_intervals = x_counterfactual.shape[-1] // window
    elif n_intervals == "sqrt":
        n_intervals = math.ceil(math.sqrt(x_counterfactual.shape[-1]))
    elif n_intervals in {"log", "log2"}:
        n_intervals = math.ceil(math.log2(x_counterfactual.shape[-1]))
    elif isinstance(n_intervals, numbers.Integral):
        if n_intervals > x_counterfactual.shape[-1]:
            raise ValueError(
                "The n_intervals parameter must be <= X_counterfactual.shape[-1]"
            )
    else:
        n_intervals = math.ceil(x_counterfactual.shape[-1] * n_intervals)

    x_counterfactual = check_array(x_counterfactual)
    x_actual = check_array(x_factual)
    y_counterfactual = check_array(
        y_counterfactual, dtype=None, ravel_1d=True, ensure_2d=False
    )

    if x_counterfactual.shape != x_actual.shape:
        raise ValueError(
            f"x_factual ({x_factual.shape}) and x_counterfactuals ({x_factual.shape}) "
            "must have the same shape."
        )

    if x_counterfactual.shape[0] != y_counterfactual.shape[0]:
        raise ValueError(
            "x_counterfactual and y_factual must have the same number of samples "
            f"{x_counterfactual.shape[0]} != {y_counterfactual.shape[0]}"
        )

    r = np.zeros(n_intervals, dtype=float)

    # indicator array of timesteps that have been changed.
    mask = ~np.isclose(x_counterfactual, x_actual)
    for i, (start, end) in enumerate(
        _intervals(x_counterfactual.shape[-1], n_intervals)
    ):
        # samples where there is a difference between the counterfactual
        # sample and the factual sample. i.e., the interval has been changed.
        idx = np.any(mask[:, start:end], axis=1)
        if np.any(idx):  # TODO: should this be all?
            x_tmp = x_counterfactual[idx, :].copy()
            x_tmp[:, start:end] = x_actual[idx, start:end]
            r[i] = (estimator.predict(x_tmp) == y_counterfactual[idx]).sum() / idx.sum()

    return r.mean() if average else r



import numpy as np
from sklearn.base import TransformerMixin
from sklearn.utils.validation import check_is_fitted

from ..base import BaseEstimator
from ..utils.validation import _check_ts_array
from ._attribute_transform import derivative_transform


class DiffTransform(TransformerMixin, BaseEstimator):
    def __init__(self, order=1):
        self.order = order

    def fit(self, X, y=None):
        self._validate_data(X)
        self.order_ = self.order
        return self

    def transform(self, X):
        self._validate_data(X, reset=False)
        check_is_fitted(self)
        return np.diff(X, n=self.order_, axis=-1)


class DerivativeTransform(TransformerMixin, BaseEstimator):
    def fit(self, X, y=None):
        self._validate_data(X, allow_3d=True)
        return self

    def transform(self, X):
        X = self._validate_data(X, reset=False)
        X_t = derivative_transform(_check_ts_array(X))

        if X.ndim == 2:
            return np.squeeze(X_t)
        else:
            return X_t



# Authors: Isak Samsten
# License: BSD 3 clause
import math
import numbers

import numpy as np
from sklearn.utils import check_scalar

from ..distance._distance import _THRESHOLD, subsequence_match
from ..distance._matrix_profile import matrix_profile
from ..utils.decorators import singleton
from ..utils.validation import check_array, check_option, check_type


@singleton
def motifs(  # noqa: PLR0912, PLR0915
    x,
    mp=None,
    *,
    window="auto",
    exclude=0.2,
    max_distance="auto",
    max_neighbours=10,
    min_neighbours=1,
    max_motif=1,
    return_distance=False,
):
    """Find motifs.

    Parameters
    ----------
    x : array-like of shape (n_samples, n_timestep)
        The time series
    mp : ndarray or shape (n_samples, profile_size), optional
        The matrix profile. The matrix profile is computed if None.
    window : "auto", int or float, optional
        The window size of the matrix profile.

        - if "auto" the window is math.ceil(0.1 * n_timesteps) if mp=None, and the
          window of the matrix profile if mp is not None.
        - if float, a fraction of n_timestep
        - if int, the exact window size
    exclude : float, optional
        The size of the exclusion zone.
    max_distance : str, optional
        The maximum distance between motifs.
    max_neighbours : int, optional
        The maximum number of neighbours
    min_neighbours : int, optional
        The minimum number of neighbours
    max_motif : int, optional
        The maximum number of motifs to return.
    return_distance : bool, optional
        Return the distance from main to neighbours

    Returns
    -------
    motif_indicies : list
        List of arrays of motif neighbour indicies
    motif_distance : list, optional
        List of arrays of distance from motif to neighbours

    See Also
    --------
    wildboar.distance.subsequence_match : find subsequence matches
    wildboar.distance.matrix_profile : compute the matrix profile

    References
    ----------
    Yeh, C. C. M. et al. (2016).
        Matrix profile I: All pairs similarity joins for time series: a unifying view
        that includes motifs, discords and shapelets. In 2016 IEEE 16th international
        conference on data mining (ICDM)
    """
    if mp is None:
        if window == "auto":
            window = 0.1

        mp = matrix_profile(x, window=window, exclude=exclude, return_index=False)
        mp = np.atleast_2d(mp)
    elif isinstance(mp, np.ndarray) and np.issubdtype(mp.dtype, np.double):
        w = x.shape[-1] - mp.shape[-1] + 1
        if window == "auto":
            window = w
        elif window != w:
            raise ValueError(
                "window == %r, expecting window == %d. Set window='auto', to "
                "automatically set window from the supplied matrix profile."
                % (window, w)
            )
        else:
            raise ValueError("window must be 'auto' or float, got %r" % window)

        mp = np.atleast_2d(mp).copy()
    else:
        raise TypeError(
            "matrix profile must be an ndarray, not %r" % type(mp).__qualname__
        )

    if max_neighbours is None:
        max_neighbours = x.shape[-1]

    if isinstance(max_distance, str):
        max_distance = check_option(_THRESHOLD, max_distance, "max_distance")

    cutoff = max_distance

    check_type(exclude, "exclude", (numbers.Integral, numbers.Real), required=False)
    if isinstance(exclude, numbers.Integral):
        exclude = check_scalar(
            exclude, "exclude", numbers.Integral, min_val=0, max_val=window
        )
    elif isinstance(exclude, numbers.Real):
        exclude = math.ceil(
            window
            * check_scalar(
                exclude,
                "exclude",
                numbers.Real,
                min_val=0,
                max_val=1,
                include_boundaries="right",
            )
        )

    x = check_array(np.atleast_2d(x), dtype=np.double)
    if x.shape[0] != mp.shape[0]:
        raise ValueError(
            "The matrix profile and x does not have the same number of samples. "
            "Set mp=None, to correctly compute the matrix profile for x."
        )

    motif_distances = []
    motif_indicies = []
    for i in range(x.shape[0]):
        motif_distance = []
        motif_index = []
        if callable(max_distance):
            cutoff = max_distance(mp[i])

        while len(motif_index) < max_motif and not np.all(np.isinf(mp)):
            candidate = np.argmin(mp[i])
            if mp[i, candidate] > cutoff:
                break

            if (
                isinstance(max_distance, numbers.Real)
                and mp[i, candidate] > max_distance
            ):
                break

            match_idx, match_dist = subsequence_match(
                x[i, candidate : candidate + window],
                x[i],
                threshold=max_distance,
                metric="scaled_euclidean",
                max_matches=max_neighbours,
                exclude=exclude,
                return_distance=True,
            )

            if match_idx.size > min_neighbours:
                motif_index.append(match_idx[:max_neighbours])
                motif_distance.append(match_dist[:max_neighbours])
                # The first match is always the same as candidate
                # so we can exclude all the matches from the matrix
                # profile
                for j in match_idx[:max_neighbours]:
                    start = max(0, j - exclude)
                    end = min(mp.shape[-1], j + exclude)
                    mp[i, start:end] = np.inf
            else:
                # Just exclude the candidate from the matrix profile
                start = max(0, candidate - exclude)
                end = min(mp.shape[-1], candidate + exclude)
                mp[i, start:end] = np.inf

        motif_distances.append(motif_distance)
        motif_indicies.append(motif_index)

    if return_distance:
        return motif_indicies, motif_distances
    else:
        return motif_indicies



# Authors: Isak Samsten
# License: BSD 3 clause
import numpy as np
from wildboar.datasets import load_two_lead_ecg
from wildboar.transform import PAA, SAX


def test_sax():
    X, _ = load_two_lead_ecg()
    X_sax = SAX(n_intervals=10, n_bins=5).fit_transform(X)

    assert X_sax.dtype == np.uint8
    np.testing.assert_equal(
        X_sax[0], np.array([3, 3, 2, 1, 0, 1, 3, 3, 3, 3], dtype=np.uint8)
    )


def test_paa():
    X, _ = load_two_lead_ecg()
    X_paa = PAA(n_intervals=10).fit_transform(X)

    assert X_paa.dtype == X.dtype
    np.testing.assert_equal(
        X_paa[0],
        np.array(
            [
                0.6462012761169009,
                0.3195832719405492,
                0.15792051772587,
                -0.6629047906026244,
                -2.3549310117959976,
                -0.39324739947915077,
                0.44239426776766777,
                0.6320434473454952,
                0.7861334010958672,
                0.30608393251895905,
            ]
        ),
    )



# Authors: Isak Samsten
# License: BSD 3 clause

"""Utilities for generating synthetic outlier datasets.

See the :ref:`User Guide
<guide-outlier-generation>` for more details and
example uses.
"""

import math
import numbers
import warnings

import numpy as np
from sklearn import clone
from sklearn.cluster import DBSCAN, OPTICS, KMeans
from sklearn.ensemble import RandomForestClassifier
from sklearn.kernel_approximation import Nystroem
from sklearn.linear_model import LogisticRegression
from sklearn.metrics import confusion_matrix
from sklearn.neighbors import NearestNeighbors
from sklearn.utils import check_random_state
from sklearn.utils.validation import _is_arraylike, check_is_fitted

from ..distance import KMedoids, pairwise_distance
from ..transform import IntervalTransform
from ..utils import _soft_dependency_error
from ..utils.validation import check_array, check_option, check_X_y

__all__ = [
    "kmeans_outliers",
    "density_outliers",
    "minority_outliers",
    "majority_outliers",
    "emmott_outliers",
]


class KernelLogisticRegression(LogisticRegression):
    """
    Kernel logistic implementation using a Nystroem kernel approximation.

    Parameters
    ----------
    kernel : str, optional
        The kernel function to use. See `sklearn.metrics.pairwise.kernel_metric`
        for kernels. The default kernel is 'rbf'.
    kernel_params : dict, optional
        Parameters to the kernel function.
    n_components : int, optional
        Number of features to construct.
    penalty : {'l1', 'l2', 'elasticnet', None}, default='l2'
        Penelty norm, for :class:`sklearn.linear_model.LogisticRegression`.
    dual : bool, default=False
        Dual or primal formulation, for
        :class:`sklearn.linear_model.LogisticRegression`.
    tol : float, default=1e-4
        Tolerance for stopping criteria.
    C : float, default=1.0
        Inverse of regularization strength; must be a positive float.
    fit_intercept : bool, default=True
        Specifies if a constant (a.k.a. bias or intercept) should be
        added to the decision function.
    intercept_scaling : float, default=1
        Intercept scaling, for :class:`sklearn.linear_model.LogisticRegression`.
    class_weight : dict or 'balanced', default=None
        Class weights, for :class:`sklearn.linear_model.LogisticRegression`.
    random_state : int, RandomState instance, default=None
        Random state, for :class:`sklearn.linear_model.LogisticRegression`.
    solver : {'lbfgs', 'liblinear', 'newton-cg', 'newton-cholesky', 'sag', 'saga'}, \
            default='lbfgs'
        Algorithm to use in the optimization problem.
    max_iter : int, default=100
        Maximum number of iterations taken for the solvers to converge.
    warm_start : bool, default=False
        When set to True, reuse the solution of the previous call to fit.
    l1_ratio : float, default=None
        The Elastic-Net mixing parameter, for
        :class:`sklearn.linear_model.LogisticRegression`.

    See Also
    --------
    wildboar.datasets.outlier.EmmottLabeler : Synthetic outlier dataset construction
    """

    def __init__(
        self,
        kernel=None,
        *,
        kernel_params=None,
        n_components=100,
        penalty="l2",
        dual=False,
        tol=1e-4,
        C=1.0,  # noqa: N803
        fit_intercept=True,
        intercept_scaling=1,
        class_weight=None,
        random_state=None,
        solver="lbfgs",
        max_iter=100,
        warm_start=False,
        l1_ratio=None,
    ):
        super().__init__(
            penalty=penalty,
            dual=dual,
            tol=tol,
            C=C,
            fit_intercept=fit_intercept,
            intercept_scaling=intercept_scaling,
            class_weight=class_weight,
            random_state=random_state,
            solver=solver,
            max_iter=max_iter,
            multi_class="auto",
            verbose=0,
            warm_start=warm_start,
            n_jobs=None,
            l1_ratio=l1_ratio,
        )
        self.kernel = kernel
        self.kernel_params = kernel_params
        self.n_components = n_components

    def fit(self, x, y, sample_weight=None):
        """
        Fit the estimator.

        Parameters
        ----------
        x : array-like of shape (n_samples, n_timestep)
            The input time series.
        y : array-like of shape (n_samples, )
            The labels.
        sample_weight : array-like of shape (n_samples, )
            The sample weights.

        Returns
        -------
        self
            The same object.
        """
        random_state = check_random_state(self.random_state)
        kernel = self.kernel or "rbf"
        n_components = min(x.shape[0], self.n_components)
        self.nystroem_ = Nystroem(
            kernel=kernel,
            kernel_params=self.kernel_params,
            n_components=n_components,
            random_state=random_state.randint(np.iinfo(np.int32).max),
        )
        self.nystroem_.fit(x)
        super().fit(self.nystroem_.transform(x), y, sample_weight=sample_weight)
        return self

    def decision_function(self, x):
        """
        Predict confidence scores for the estimator.

        Parameters
        ----------
        x : array-like of shape (n_samples, n_timestep)
            The input time series.

        Returns
        -------
        ndarray of shape (n_samples, )
            The confidence scores.
        """
        check_is_fitted(self)
        return super().decision_function(self.nystroem_.transform(x))


def kmeans_outliers(
    x,
    *,
    n_outliers=0.05,
    n_clusters=5,
    random_state=None,
):
    """
    K-mean outlier generation.

    Label the samples of the cluster farthers from the other clusters as
    outliers.

    Parameters
    ----------
    x : ndarray of shape (n_samples, n_timestep)
        The input samples.
    n_outliers : float, optional
        The number of outlier samples expressed as a fraction of the inlier samples.

        - if float, the number of outliers are guaranteed but an error is raised
          if no cluster can satisfy the constraints. Lowering the `n_cluster`
          parameter to allow for more samples per cluster.
    n_clusters : int, optional
        The number of clusters.
    random_state : int or RandomState
        - If `int`, `random_state` is the seed used by the random number generator
        - If `RandomState` instance, `random_state` is the random number generator
        - If `None`, the random number generator is the `RandomState` instance used
            by `np.random`.

    Returns
    -------
    x_outlier : ndarray of shape (n_inliers + n_outliers, n_timestep)
        The samples.
    y_outlier : ndarray of shape (n_inliers + n_outliers, )
        The inliers (labeled as 1) and outlier (labled as -1).
    """
    x = check_array(x)
    random_state = check_random_state(random_state)
    k_means_ = KMeans(
        n_init=10,  # TODO(1.4) Move to "auto"
        n_clusters=n_clusters,
        random_state=random_state.randint(np.iinfo(np.int32).max),
    )

    k_means_.fit(x)
    centroid_distance = pairwise_distance(k_means_.cluster_centers_, metric="euclidean")

    # skip self matches and invalid clusters
    centroid_distance[centroid_distance == 0] = np.nan

    if np.all(np.isnan(centroid_distance)):
        raise ValueError("There are no valid clusters.")

    # hide the warning for ignored clusters
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RuntimeWarning)
        outlier_cluster_ = np.nanargmax(np.nanmean(centroid_distance, axis=1))

    return _make_outlier_arrays(
        x,
        n_outliers=n_outliers,
        outlier_indicies=np.where(k_means_.labels_ == outlier_cluster_)[0],
        inlier_indicies=np.where(k_means_.labels_ != outlier_cluster_)[0],
        random_state=random_state,
    )


def density_outliers(
    x,
    *,
    n_outliers=0.05,
    method="dbscan",
    eps=2.0,
    min_sample=5,
    metric="euclidean",
    max_eps=np.inf,
    random_state=None,
):
    """
    Densitiy based outlier generation.

    Labels samples as outliers if a density cluster algorithm fail to assign
    them to a cluster.

    Parameters
    ----------
    x : ndarray of shape (n_samples, n_timestep)
        The input samples.
    n_outliers : float, optional
        The number of outlier samples expressed as a fraction of the inlier samples.
        By default, all samples of the minority class is considered as outliers.
    method : {"dbscan", "optics"}, optional
        The density based clustering method.
    eps : float, optional
        The eps parameter, when `method="dbscan"`.
    min_sample : int, optional
        The `min_sample` parameter to the cluter method.
    metric : str, optional
        The `metric` parameter to the cluster method.
    max_eps : float, optional
        The `max_eps` parameter, when `method="optics"`.
    random_state : int or RandomState
        - If `int`, `random_state` is the seed used by the random number generator
        - If `RandomState` instance, `random_state` is the random number generator
        - If `None`, the random number generator is the `RandomState` instance used
            by `np.random`.

    Returns
    -------
    x_outlier : ndarray of shape (n_inliers + n_outliers, n_timestep)
        The samples.

    y_outlier : ndarray of shape (n_inliers + n_outliers, )
        The inliers (labeled as 1) and outlier (labled as -1).
    """
    if method == "dbscan":
        estimator = DBSCAN(eps=eps, min_samples=min_sample, metric=metric)
    elif method == "optics":
        estimator = OPTICS(max_eps=max_eps, min_samples=min_sample, metric=metric)
    else:
        raise ValueError("method must be 'dbscan' or 'optics', got %s." % method)

    estimator.fit(x)
    label, count = np.unique(estimator.labels_, return_counts=True)
    if len(count) == 1:
        raise ValueError("A single cluster was formed.")
    elif not np.any(label == -1):
        raise ValueError("There are no outlier points.")

    return _make_outlier_arrays(
        x,
        n_outliers=n_outliers,
        outlier_indicies=np.where(estimator.labels_ == -1)[0],
        inlier_indicies=np.where(estimator.labels_ != -1)[0],
        random_state=check_random_state(random_state),
    )


def majority_outliers(x, y, *, n_outliers=0.05, random_state=None):
    """
    Label the majority class as inliers.

    Parameters
    ----------
    x : ndarray of shape (n_samples, n_timestep)
        The input samples.
    y : ndarray of shape (n_samples, )
        The input labels.
    n_outliers : float, optional
        The number of outlier samples expressed as a fraction of the inlier samples.
    random_state : int or RandomState
        - If `int`, `random_state` is the seed used by the random number generator
        - If `RandomState` instance, `random_state` is the random number generator
        - If `None`, the random number generator is the `RandomState` instance used
            by `np.random`.

    Returns
    -------
    x_outlier : ndarray of shape (n_inliers + n_outliers, n_timestep)
        The samples.
    y_outlier : ndarray of shape (n_inliers + n_outliers, )
        The inliers (labeled as 1) and outlier (labled as -1).
    """
    x = check_array(x, allow_3d=True)
    y = check_array(y, ensure_2d=False)
    labels, counts = np.unique(y, return_counts=True)
    if len(labels) < 2:
        raise ValueError("Two labels are required, got %r" % len(labels))

    outlier_indicator = np.isin(y, labels[labels != labels[np.argmax(counts)]])
    return _make_outlier_arrays(
        x,
        n_outliers=n_outliers,
        outlier_indicies=outlier_indicator.nonzero()[0],
        inlier_indicies=(~outlier_indicator).nonzero()[0],
        random_state=check_random_state(random_state),
    )


def minority_outliers(x, y, *, n_outliers=0.05, random_state=None):
    """
    Label (a fraction of) the minority class as the outlier.

    Parameters
    ----------
    x : ndarray of shape (n_samples, n_timestep)
        The input samples.
    y : ndarray of shape (n_samples, )
        The input labels.
    n_outliers : float, optional
        The number of outlier samples expressed as a fraction of the inlier samples.

        - if float, the number of outliers are guaranteed but an error is raised
          if the minority class has to few samples.
    random_state : int or RandomState
        - If `int`, `random_state` is the seed used by the random number generator
        - If `RandomState` instance, `random_state` is the random number generator
        - If `None`, the random number generator is the `RandomState` instance used
            by `np.random`.

    Returns
    -------
    x_outlier : ndarray of shape (n_inliers + n_outliers, n_timestep)
        The samples.
    y_outlier : ndarray of shape (n_inliers + n_outliers, )
        The inliers (labeled as 1) and outlier (labled as -1).
    """
    x = check_array(x, allow_3d=True)
    y = check_array(y, ensure_2d=False)

    labels, label_count = np.unique(y, return_counts=True)
    if len(labels) < 2:
        raise ValueError("Two labels are required, got %r" % len(labels))

    outlier_label = labels[np.argmin(label_count)]
    return _make_outlier_arrays(
        x,
        n_outliers=n_outliers,
        outlier_indicies=np.where(y == outlier_label)[0],
        inlier_indicies=np.where(y != outlier_label)[0],
        random_state=check_random_state(random_state),
    )


def emmott_outliers(  # noqa: PLR0912, PLR0915
    x,
    y,
    *,
    n_outliers=None,
    confusion_estimator=None,
    difficulty_estimator=None,
    transform="interval",
    difficulty="simplest",
    scale=None,
    variation="tight",
    random_state=None,
):
    """
    Difficulty based outlier generation.

    Create a synthetic outlier detection dataset from a labeled classification
    dataset using the method described by Emmott et.al. (2013).

    The Emmott labeler can reliably label both binary and multiclass datasets. For
    binary datasets a random label is selected as the outlier class. For multiclass
    datasets a set of classes with maximal confusion (as measured by
    `confusion_estimator` is selected as outlier label. For each outlier sample the
    `difficulty_estimator` assigns a difficulty score which is digitized into ranges
    and selected according to the `difficulty` parameters. Finally a sample of
    approximately `n_outlier` is selected either maximally dispersed or tight.

    Parameters
    ----------
    x : ndarray of shape (n_samples, n_timestep)
        The input samples.
    y : ndarray of shape (n_samples, )
        The input labels.
    n_outliers : float, optional
        The number of outlier samples expressed as a fraction of the inlier samples.

        - if float, the number of outliers are guaranteed but an error is raised
          if the the requested difficulty has to few samples or the labels selected
          for the outlier label has to few samples.
    confusion_estimator : object, optional
        Estimator of class confusion for datasets where `n_classes > 2`. Default to a
        random forest classifier.
    difficulty_estimator : object, optional
        Estimator for sample difficulty. The difficulty estimator must support
        `predict_proba`. Defaults to a kernel logistic regression model with
        a RBF-kernel.
    transform : 'interval' or Transform, optional
        Transform x before the confusion and difficulty estimator.

        - if None, no transformation is applied.
        - if 'interval', use the :class:`transform.IntervalTransform` with default
          parameters.
        - otherwise, use the supplied transform.
    difficulty : {'any', 'simplest', 'hardest'}, int or array-like, optional
        The difficulty of the outlier points quantized according to scale. The value
        should be in the range `[1, len(scale)]` with lower difficulty denoting
        simpler outliers. If an array is given, multiple difficulties can be
        included, e.g., `[1, 4]` would mix easy and difficult outliers.

        - if 'any' outliers are sampled from all scores.
        - if 'simplest' the simplest n_outliers are selected.
        - if 'hardest' the hardest n_outliers are selected.
    scale : int or array-like, optional
        The scale of quantized difficulty scores. Defaults to `[0, 0.16, 0.3, 0.5]`.
        Scores (which are probabilities in the range [0, 1]) are fit into the ranges
        using `np.digitize(difficulty, scale)`.

        - if int, use `scale` percentiles based in the difficulty scores.
    variation : {'tight', 'dispersed'}, optional
        Selection procedure for sampling outlier samples. If `difficulty="simplest"`
        or `difficulty="hardest"`, this parameter has no effect.

        - if 'tight' a pivot point is selected and the `n_outlier` closest samples
          are selected according to their euclidean distance.
        - if 'dispersed' `n_outlier` points are selected according to a facility
          location algorithm such that they are distributed among the outliers.
    random_state : int or RandomState
        - If `int`, `random_state` is the seed used by the random number generator
        - If `RandomState` instance, `random_state` is the random number generator
        - If `None`, the random number generator is the `RandomState` instance used
            by `np.random`.

    Returns
    -------
    x_outlier : ndarray of shape (n_inliers + n_outliers, n_timestep)
        The samples.
    y_outlier : ndarray of shape (n_inliers + n_outliers, )
        The inliers (labeled as 1) and outlier (labled as -1).

    Warnings
    --------
    n_outliers
        The number of outliers returned is dependent on the difficulty setting and the
        available number of samples of the minority class. If the minority class does
        not contain sufficient number of samples of the desired difficulty, fewer than
        n_outliers may be returned.

    Notes
    -----
    - For multiclass datasets the Emmott labeler require the package `networkx`

    The difficulty parameters 'simplest' and 'hardest' are not described by
    Emmott et.al. (2013)

    References
    ----------
    Emmott, A. F., Das, S., Dietterich, T., Fern, A., & Wong, W. K. (2013).
        Systematic construction of anomaly detection benchmarks from real data.
        In Proceedings of the ACM SIGKDD workshop on outlier detection and description
        (pp. 16-21).
    """
    x, y = check_X_y(x, y, dtype=float)
    random_state = check_random_state(random_state)
    n_classes = np.unique(y).shape[0]

    if transform == "interval":
        transform = IntervalTransform()
    elif transform is not None:
        transform = clone(transform)

    if transform is not None:
        transform.fit(x)

    if n_classes > 2:
        if confusion_estimator is None:
            confusion_estimator = RandomForestClassifier(n_jobs=-1, oob_score=True)
        else:
            confusion_estimator = clone(confusion_estimator)

        _set_random_states(
            confusion_estimator, random_state.randint(np.iinfo(np.int32).max)
        )
        outlier_label = _emmott_multiclass_outlier_class(
            x,
            y,
            confusion_estimator=confusion_estimator,
            transform=transform,
            n_labels=n_classes,
        )
    elif n_classes == 2:
        labels, counts = np.unique(y, return_counts=True)
        outlier_label = labels[np.argmin(counts)]
    else:
        raise ValueError("Two labels are required, got %r" % n_classes)

    outlier_indicator = np.isin(y, outlier_label)
    y_new = np.ones(x.shape[0], dtype=int)
    y_new[outlier_indicator] = -1

    if difficulty_estimator is None:
        difficulty_estimator = KernelLogisticRegression(
            kernel="poly",
            max_iter=1000,
            random_state=random_state.randint(np.iinfo(np.int32).max),
        )
    else:
        difficulty_estimator = clone(difficulty_estimator)
        _set_random_states(
            difficulty_estimator, random_state.randint(np.iinfo(np.int32).max)
        )

    outliers_indices = np.where(y_new == -1)[0]
    inlier_indices = np.where(y_new == 1)[0]
    x_outliers = x[outliers_indices]
    y_outliers = y_new[outliers_indices]

    difficulty_estimate = _emmott_estimate_difficulty(
        x, y_new, difficulty_estimator=difficulty_estimator, transform=transform
    )
    difficulty_estimate = difficulty_estimate[outliers_indices].reshape(-1)

    outlier_fraction = y_outliers.shape[0] / x.shape[0]
    if outlier_fraction > n_outliers:
        n_outliers_ = min(
            y_outliers.shape[0], math.ceil(n_outliers * inlier_indices.size)
        )

        if isinstance(difficulty, str):
            if difficulty == "any":
                outlier_selector = np.arange(0, difficulty_estimate.shape[0])
            elif difficulty == "simplest":
                outlier_selector = np.argpartition(difficulty_estimate, n_outliers_)[
                    :n_outliers_
                ]
            elif difficulty == "hardest":
                outlier_selector = np.argpartition(difficulty_estimate, -n_outliers_)[
                    -n_outliers_:
                ]
            else:
                raise ValueError(
                    "difficulty must be 'any', 'simplest' or 'hardest', got %r."
                    % difficulty
                )
        else:
            if scale is None:
                scale = _DEFAULT_EMMOTT_SCALE
            elif isinstance(scale, numbers.Integral):
                scale = np.percentile(
                    difficulty_estimate, np.linspace(0, 100, scale, endpoint=False)
                )
            elif _is_arraylike(scale):
                scale = np.array(scale)
            else:
                raise ValueError(
                    "scale must be int or array-like, not %r" % type(scale).__qualname__
                )
            difficulty_scores = np.digitize(difficulty_estimate, scale)
            difficulties = np.unique(difficulty_scores)
            if difficulty not in difficulties:
                options = ", ".join([str(d) for d in difficulties[:-1]])
                raise ValueError(
                    f"difficulty must be {options} or {difficulties[-1]}, "
                    f"got {difficulty}"
                )
            outlier_selector = np.isin(difficulty_scores, difficulty)

        x_outliers = x_outliers[outlier_selector]
        variation = check_option(_EMMOTT_VARIATION, variation, "variation")
        outlier_sampled = variation(
            x_outliers, n_outliers_, random_state.randint(np.iinfo(np.int32).max)
        )
    else:
        outlier_sampled = np.arange(y_outliers.shape[0])

    if outlier_sampled.shape[0] / x.shape[0] < n_outliers:
        n_inliers = math.ceil(
            outlier_sampled.shape[0] / n_outliers - outlier_sampled.shape[0]
        )
        random_state.shuffle(inlier_indices)
        inlier_indices = inlier_indices[:n_inliers]

    return (
        np.concatenate([x[inlier_indices], x_outliers[outlier_sampled]], axis=0),
        np.concatenate([y_new[inlier_indices], y_outliers[outlier_sampled]], axis=0),
    )


def _make_outlier_arrays(
    x, *, n_outliers, outlier_indicies, inlier_indicies, random_state
):
    random_state.shuffle(outlier_indicies)
    random_state.shuffle(inlier_indicies)
    outlier_fraction = outlier_indicies.size / x.shape[0]
    if outlier_fraction < n_outliers:
        n_inliers = math.ceil(
            outlier_indicies.size / n_outliers - outlier_indicies.size
        )
        n_outliers = outlier_indicies.size
    elif outlier_fraction > n_outliers:
        n_outliers = min(
            outlier_indicies.size, math.ceil(n_outliers * inlier_indicies.size)
        )
        n_inliers = inlier_indicies.size
    else:
        n_outliers = outlier_indicies.size
        n_inliers = inlier_indicies.size

    x_outlier = x[outlier_indicies[:n_outliers]]
    x_inlier = x[inlier_indicies[:n_inliers]]
    x = np.concatenate([x_outlier, x_inlier], axis=0)
    y = np.ones(x.shape[0])
    y[: x_outlier.shape[0]] = -1
    return x, y


def _emmott_estimate_difficulty(
    x,
    y_new,
    *,
    difficulty_estimator,
    transform=None,
):
    if transform is not None:
        x = transform.transform(x)

    difficulty_estimator.fit(x, y_new)
    if hasattr(difficulty_estimator, "oob_decision_function_"):
        difficulty_estimate = difficulty_estimator.oob_decision_function_
    else:
        difficulty_estimate = difficulty_estimator.predict_proba(x)

    difficulty_estimate = difficulty_estimate[
        :, np.where(difficulty_estimator.classes_ == 1)[0]
    ]

    return difficulty_estimate


def _emmott_multiclass_outlier_class(x, y, *, confusion_estimator, transform, n_labels):
    if transform is not None:
        transform.transform(x)

    confusion_estimator.fit(x, y)
    try:
        import networkx as nx

        if hasattr(confusion_estimator, "oob_decision_function_"):
            y_pred = confusion_estimator.classes_[
                np.argmax(confusion_estimator.oob_decision_function_, axis=1)
            ]
        else:
            y_pred = confusion_estimator.predict(transform.transform(x))

        cm = confusion_matrix(y, y_pred)  # TODO: use probabilities
        graph = nx.Graph()
        labels = confusion_estimator.classes_
        graph.add_nodes_from(labels)
        for i in range(n_labels):
            for j in range(n_labels):
                if i != j:
                    graph.add_edge(labels[i], labels[j], weight=cm[i][j] + cm[j][i])

        max_spanning_tree = nx.maximum_spanning_tree(graph, algorithm="kruskal")
        coloring = nx.algorithms.bipartite.color(max_spanning_tree)
        labeling = {1: [], 0: []}
        for label in labels:
            labeling[coloring[label]].append(label)

        zero = np.isin(y, labeling[0])
        one = np.isin(y, labeling[1])

        if np.sum(zero) <= np.sum(one):
            outlier_label = np.array(labeling[0])
        else:
            outlier_label = np.array(labeling[1])
    except ModuleNotFoundError as e:
        _soft_dependency_error(e, package="networkx", context="emmott_outlier")

    return outlier_label


def _variation_dispersed(x, n_outliers, random_state):
    """
    Find the indices of the medoids for a given dataset using the KMedoids algorithm.

    Parameters
    ----------
    x : array-like
        The input dataset to find the medoids for.
    n_outliers : int
        The number of medoids to find. This value should be less than or equal
        to the number of samples in the dataset.

    random_state : int or RandomState instance, optional (default=None)
        Determines random number generation for centroid initialization. Use an
        int to make the randomness deterministic. See
        :class:`~numpy.random.RandomState` for details.

    Returns
    -------
    array-like
        The indices of the medoids in the input dataset.
    """
    n_outliers = min(x.shape[0], n_outliers)
    f = KMedoids(n_clusters=n_outliers, random_state=random_state).fit(x)
    return f.medoid_indices_


def _variation_tight(x, n_outliers, random_state):
    random_state = check_random_state(random_state)
    n_outliers = min(x.shape[0], n_outliers)
    if n_outliers == x.shape[0]:
        return np.arange(0, x.shape[0])
    sample = random_state.randint(0, x.shape[0])
    f = NearestNeighbors(n_neighbors=n_outliers).fit(x)
    return f.kneighbors(x[sample].reshape(1, -1), return_distance=False).reshape(-1)


def _set_random_states(estimator, random_state=None):
    random_state = check_random_state(random_state)
    to_set = {}
    for key in sorted(estimator.get_params(deep=True)):
        if key == "random_state" or key.endswith("__random_state"):
            to_set[key] = random_state.randint(np.iinfo(np.int32).max)

    if to_set:
        estimator.set_params(**to_set)


_EMMOTT_VARIATION = {"tight": _variation_tight, "dispersed": _variation_dispersed}
_DEFAULT_EMMOTT_SCALE = np.array([0, 0.16, 0.3, 0.5])



import numpy as np
import pytest
from numpy.testing import assert_almost_equal

from wildboar.distance._multi_metric import make_metric, make_metrics


def test_make_metric_missing_metric():
    with pytest.raises(ValueError):
        make_metric("missing_metric")


def test_make_metric_missing_parameter():
    with pytest.raises(TypeError):
        make_metric("dtw", max_missing_metric=1, min_missing_metric=0)


def test_make_metric_invalid_spec():
    with pytest.raises(ValueError, match="The parameter invalid_spec"):
        make_metric("dtw", invalid_spec=0)


def test_make_metric_missing_min_max():
    with pytest.raises(ValueError, match="The maximum"):
        make_metric("dtw", min_r=0.0)

    with pytest.raises(ValueError, match="The minimum"):
        make_metric("dtw", max_r=0.0)


def test_make_metric_single_spec():
    metrics = make_metric("dtw", min_r=0.0, max_r=0.25, default_n=10)
    assert len(metrics) == 10


def test_make_metric_muli_spec():
    metrics = make_metric(
        "wdtw", min_r=0.0, max_r=0.25, min_g=0.01, max_g=1, default_n=10
    )
    assert len(metrics) == 10 * 10


def test_make_metric_multi_spec_different_n():
    metrics = make_metric(
        "wdtw", num_r=3, min_r=0.0, max_r=0.25, min_g=0.01, max_g=1, default_n=10
    )
    assert len(metrics) == 3 * 10


def test_make_metrics():
    spec = [
        ("euclidean", None),
        ("dtw", None),
        ("dtw", dict(min_r=0.0, max_r=0.25, num_r=10)),
    ]

    metrics, weights = make_metrics(spec)

    assert len(metrics) == 1 + 1 + 10
    assert_almost_equal(
        weights,
        np.array(
            [
                0.333333333,
                0.333333333,
                0.0333333333,
                0.0333333333,
                0.0333333333,
                0.0333333333,
                0.0333333333,
                0.0333333333,
                0.0333333333,
                0.0333333333,
                0.0333333333,
                0.0333333333,
            ]
        ),
    )
